/*
 * Copyright (c) 1993-2007 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <portLib.h>
#include <h2mathLib.h>

#include "genBasic/genBasicStruct.h"
#include "platineCartesian.h"

#ifdef MAIN_TO_TEST
int main(int argc, char **argv) 
{
  GEN_POINT_3D goal;
  GEN_POINT_3D camPos;
  double yaw, pitch, dist;

  printf ("x: "); fflush(stdout);
  fscanf(stdin, "%lf", &goal.x);
  printf ("y: "); fflush(stdout);
  fscanf(stdin, "%lf", &goal.y);
  printf ("z: "); fflush(stdout);
  fscanf(stdin, "%lf", &goal.z);

  camPos.x = 0.0;
  camPos.y = 0.0;
  camPos.z = 0.0;

  platineCartToPolar(&goal, &camPos, 
		     &yaw, &pitch, &dist);

  return 0;
}
#endif /* MAIN_TO_TEST */


/* platineCartToPolar -

   Computation of yaw and pitch of the PTU in order to centralize 
   in the camera frame (yc=zc=0) an object given in the PTU-base 
   cartesian frame (similar to robot frame but a translation).

  WARNING: the camera frame must be parallel to the PTU frame !!

  Full 3D equation:
  -----------------
  T platine_rot/platine_base * T cam/platine_rot * Obj/cam = Obj/platine_base

                                (cy -sy 0 0) ( cp 0 sp 0)
  T platine_rot/platine_base =  (sy  cy 0 0) ( 0  1  0 0)
                                (0   0  1 0) (-sp 0 cp 0)
                                (0   0  0 1) ( 0  0  0 1)

   ( cy cp   -sy    cy sp   0 )  (R0  R1  R2  T0)  (d)      (x)
   ( sy cp    cy    sy sp   0 )  (R3  R4  R5  T1)  (0)   =  (y)
   (  -sp     0       cp    0 )  (R6  R7  R8  T2)  (0)      (z)
   (    0     0       0     1 )  (0    0   0   1)  (0)      (1)

   | sy x - cy y = -(d R3 + T1)
   | cy x + sy y =   cp (d R0 + T0) + sp (d R6 + T2)
   |           z = - sp (d R0 + T0) + cp (d R6 + T2)

   Thus:
   | x  sy - y cy = -(d R3 + T1)
   | x' sp + z cp = dR6+T2
   | x' cp - z sp = dR0+T0
   with: x' = x cy + y sy


   Reduced equation (simply the resolution)
   ----------------
   Hypothesis : cam  parallel to platine_rot
         thus : R0=R4=R8=1  and R1=R2=R3=R5=R6=R7=0

   Thus: 
   (1)    x  sy - y cy = -T1
   (2)    x' sp + z cp =  T2
   (3)    x' cp - z sp =  d + T0

   With: sy = sqrt(1-cy^2)
   then :  (1) =>  cy^2 (x^2 + y^2) - 2 T1 y cy + T1^2-x^2 = 0

   Second degre equation resolution:
      det' = b'^2 -ac
      cy = (-b' + sqrt(det')) / a

   (2) => Similar for yaw but remplace:
   x by x' (= x cy + y sy) ; y by -z; T1 by -T2

   Finaly (3) gives d  (just to check coherency)
*/

BOOL platineCartToPolar(GEN_POINT_3D *goal, GEN_POINT_3D *camPos, 
			double *yaw, double *pitch, double *dist)
{
  double dist2, det, cPitch, cYaw, xy;

  /* compute yaw */
  dist2 = SQR(goal->x) + SQR(goal->y);
  det = (dist2 - SQR(camPos->y)) * SQR(goal->x);

  if (det < 0) {
    printf ("platineCartToPolar: negative determinant for yaw dist2 = %g det = %g  x=%g\n", dist2, det, goal->x);
    return FALSE;
  }
  cYaw = (camPos->y * goal->y + sqrt(det)) / dist2;

  *yaw = acos(cYaw);
  /* target behind the camera */
  if (goal->x < camPos->x)
    *yaw = *yaw + 3.1415926/2.0;
  /* target on right side of the robot */
  if (goal->y < camPos->y)
    *yaw = -(*yaw);

#if 0
  fprintf (stderr, "yaw %5.2f: dist2 = %g det = %g  cP=%g  sP =%g\n", 
		 *yaw * 180.0 / 3.1415926,
		 dist2, det, cYaw, sin(*yaw));
#endif

  /* compute pitch */
  xy = cYaw * goal->x + sin(*yaw) * goal->y;
  dist2 = SQR(xy) + SQR(goal->z);
  det = (dist2 - SQR(camPos->z)) * SQR(xy);
  if (det < 0) {
    printf ("platineCartToPolar: negative determinant for pitch\n");
    return FALSE;
  }
  cPitch = (camPos->z * goal->z + sqrt(det)) / dist2;

  *pitch = acos(cPitch);

  if (goal->z > camPos->z)
    *pitch = -(*pitch);

#if 0
  fprintf (stderr, "pitch   %5.2f: dist2 = %g det = %g xy = %g  cos=%g\n", 
		 *pitch * 180.0 / 3.1415926,
		 dist2, det, xy, cPitch);
#endif  

  /* compute dist */
  *dist =  cPitch * xy - sin(*pitch) * goal->z - camPos->x;

#if 0
  fprintf (stderr, "yaw = %5.2f pitch = %5.2f dist = %5.2f\n", 
		 *yaw * 180.0 / 3.1415926,
		 *pitch * 180.0 / 3.1415926,
		 *dist);
#endif
  return TRUE;
}
