/*
 * Copyright (C) 2000-2005 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include <portLib.h>
#include <taskLib.h>
#include <sysLib.h>

#include "h2mathLib.h"

#include "ptDriver.h"
#include "ptDriverProto.h"
#include "ptDriverPrivate.h"
#include "ptClientProto.h"

/* ------------------------- EXTERNAL CONSTANTS ------------------------ */

extern const PT_CMD_DESC ptCmdDesc[];

extern const PT_ERR_FORMAT ptErrDesc[];

extern const PT_STRINGS_FORMAT ptStringsDesc[];


/* -------------------------- LOCAL CONSTANTS  ------------------------- */

static PT_DRIVER_STR *ptFirstDriverSDI=NULL;
#define PT_NB_MAX_DRIVERS 6
static PT_DRIVER_STR *ptDriversTab[PT_NB_MAX_DRIVERS] = {NULL};


static const PT_CMD platineInitGetConst[] = PT_INIT_GET_CONST;

const char *ptCmdStatusStrings[] = PT_CMD_STATUS_STRINGS;

/* ------------------ LOCAL FUNCTIONS PROTOTYPES  --------------------- */

static BOOL ptSendInitCmd(PT_DRIVER_STR *ptDriverSDI);
static PT_DRIVER_STR * ptAllocSDI(BOOL verbose, char *ptDriverName);
static void ptInitSDI(PT_DRIVER_STR *ptDriverSDI);
static void ptFreeSDI(PT_DRIVER_STR *ptDriverSDI);
static PT_DRIVER_STR * ptLookForDriver(char *ptDriverName);
static PT_CMD_STATUS ptGetCmdStatus (int cmd, PT_DRIVER_STR *ptDriverSDI);
static PT_CMD_STATUS ptGetAndUpdateCmdStatus (int cmd, PT_DRIVER_STR *ptDriverSDI);
static PT_CMD_STATUS ptGetAndUpdateCmdStatusNoProtect (int cmd, 
						       PT_DRIVER_STR *ptDriverSDI);
static BOOL ptTestCmd(int cmd, PT_DRIVER_STR *ptDriverSDI);
static BOOL ptSendAsciiCmd(char *cmd, PT_DRIVER_STR *ptDriverSDI);
static PT_STRINGS_DECODE ptDecodeString(char *string, PT_DRIVER_STR *ptDriverSDI);
static void ptSpyCmdInput(int cmd, PT_DRIVER_STR *ptDriverSDI);

static BOOL ptAddWaitedReplies(PT_CMD cmd, PT_DRIVER_STR *ptDriverSDI);

/* ====================================================================== *
 *                   DRIVER and CLIENT INITIALISATIONS                    *
 *     (to be done : should be separate to allow several clients ?)       *
 * ====================================================================== */

/* 
 * ptInitDriver  -  starts driver, allocates driver and client structures
 *                  and sends init cmds
 */
PT_DRIVER_STR *
ptInitDriver(char *device, int baudRate, int verbose)
{
  PT_DRIVER_STR *ptDriverSDI=NULL;
  PT_DRIVER_ARGS *args;
  char ptDriverName[32];
  int i=0;
  char c[]="c";
  int timeout = 10*sysClkRateGet();  /* 10 * nb tics / sec */
  int taskId;

  /* driver name */
  if (device==NULL || device[0] == '\0') {
    printf ("ptInitDriver: unvalid device name\n");
    return NULL;
  }
  strcpy(ptDriverName, "ptDriver");
  do {
    c[0] = device[i++];
    if (c[0] == '\0')
      break;
    /* removes '/'  */
    if (c[0] != '/') {
      strcat (ptDriverName, c);
      continue;
    }
    if (strlen(ptDriverName) >= PT_DRIVER_NAME_LENGTH) {
      printf ("ptInitDriver: warning : device name too long\n");
      break;
    }
  } while (1);
  
  /* Is the driver running already ? */
  if ((ptDriverSDI = ptLookForDriver(ptDriverName))) {

    printf ("ptInitDriver: driver %s already running, just reset it\n",
	    ptDriverName);

   ptInitSDI(ptDriverSDI);
  }
  
  /* Start the driver */
  else {

    /* Allocation of the driver data structure */
    if ((ptDriverSDI = ptAllocSDI(verbose,ptDriverName))==NULL)
      return NULL;
    
    /* Starts the driver process */
    args = (PT_DRIVER_ARGS *)malloc(sizeof(PT_DRIVER_ARGS));
    args->device = device;
    args->baudRate = baudRate;
    args->sdiPtr = ptDriverSDI;
    taskId = taskSpawn2 (ptDriverName, 5, VX_FP_TASK, 5000, ptDriver, args);
    if (taskId == -1) {
      printf ("platineInitDriver: cannot create driver task\n");
      return NULL;
    }      
    ptDriverSDI->privateData->taskId = taskId;

    /* Waits end of init */
    while (ptDriverSDI->ptStopDriverFlag && timeout) {
      timeout --;
      taskDelay(10);
    }
    
    if (timeout==0) {
      printf ("platineInitDriver: cannot start driver (timeout)\n");
      return NULL;
    }
  }

  /* Sends init commands */
  if (!ptSendInitCmd(ptDriverSDI)) {
    fprintf (stderr, "ptSendInitCmd failed: \n");
/*     taskDelete(ptDriverSDI->privateData->taskId); */
/*     ptDriverStop(ptDriverSDI); */
    return NULL;
  }
  ptDriverSDI->ptDriverInitialized = TRUE;

  return ptDriverSDI;
}

/*
 *  ptSendInitCmd  -  Sends commands to get the constants of the PTU
 *                    (max speeds, angle bounds, ...). The list of the 
 *                    commands is given by platineInitGetConst in
 *                    ptDriverPrivate.h.
 *
 *  Returns : TRUE if init OK, or FALSE if not.
 */

static BOOL ptSendInitCmd(PT_DRIVER_STR *ptDriverSDI)
{
  int i;
  double val;
  int nbInitCmd;
  PT_ERR err;
  char version[PT_DRIVER_NAME_LENGTH];
  int v0, v1, v2;


  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Disables echo mode */
  ptSendAsciiCmd ("ed", ptDriverSDI);
  ptSendAsciiCmd ("ed", ptDriverSDI);
/*   if (!ptSendAndGetPrint(PT_ECHO_OFF, 0, ptDriverSDI))  */
/*     return FALSE;  */

  /* firmware controler version */
  if (!ptSendAndGetPrint(PT_VERSION, 0, ptDriverSDI)) 
    return FALSE;  
  ptDriverSDI->privateData->ptCmdState[PT_VERSION].replyStatus = PT_CONST_OK;
  ptGetString(PT_VERSION, version, &err, ptDriverSDI);
  if (sscanf(version, "v%d.%d.%d", &v0, &v1, &v2) != 3) {
    fprintf (stderr, "ptDriver: cannot get version. Uses default one.\n");
  }
  ptUpdateVersion(ptDriverSDI, v0, v1, v2);

  /* Get pan axis resolution */
  if (!ptSendAndGetPrint(GET_RES_PAN, 0, ptDriverSDI)) 
    return FALSE;
  ptDriverSDI->privateData->ptCmdState[GET_RES_PAN].replyStatus = PT_CONST_OK;
  ptGetDouble(GET_RES_PAN, &ptDriverSDI->ptResolution, &err, ptDriverSDI);

  /* Get tilt axis resolution */
  if (!ptSendAndGetPrint(GET_RES_TILT, 0, ptDriverSDI)) 
    return FALSE;
  ptDriverSDI->privateData->ptCmdState[GET_RES_TILT].replyStatus = PT_CONST_OK;
  ptGetDouble(GET_RES_TILT, &val, &err, ptDriverSDI);

  /* Checks resolution coherency */
  if (val != ptDriverSDI->ptResolution) {
    fprintf (stderr, "ptDriver: different resolutions for pan (%f)"
	     " and tilt (%f) not allowed !\n", ptDriverSDI->ptResolution, val);
    return FALSE;
  }
  ptDriverSDI->ptDeg2step = 3600/ptDriverSDI->ptResolution;
  ptDriverSDI->ptStep2deg = ptDriverSDI->ptResolution/3600;

  ptDriverSDI->ptRad2step = RAD_TO_DEG(ptDriverSDI->ptDeg2step);
  ptDriverSDI->ptStep2rad = DEG_TO_RAD(ptDriverSDI->ptStep2deg);


  /* Gets the list of commands to get all PTU constants */
  nbInitCmd = sizeof(platineInitGetConst)/sizeof(PT_CMD);

  /* Sends every commands */
  for (i = 0; i < nbInitCmd; i++) {
    if (!ptSendAndGetPrint(platineInitGetConst[i], 0, ptDriverSDI)) 
      return FALSE;
    ptDriverSDI->privateData->ptCmdState[platineInitGetConst[i]].replyStatus = PT_CONST_OK;
  }

  ptDriverSDI->ptDriverInitialized = TRUE;

  return TRUE;
}

static PT_DRIVER_STR *
ptAllocSDI(BOOL verbose, char *ptDriverName)
{
  PT_DRIVER_STR *ptDriverSDI;
  int i;

  /* On alloue la structure PT_DRIVER_STR  */
  ptDriverSDI = (PT_DRIVER_STR *)malloc(sizeof(PT_DRIVER_STR));
  if (ptDriverSDI==NULL) {
    printf("ptAllocSDI: alloc ptDriverSDI failed\n");
    return NULL;
  }
  memset(ptDriverSDI, 0, sizeof(PT_DRIVER_STR));

  ptDriverSDI->privateData = (PT_DRIVER_PRIVATE *)malloc(sizeof(PT_DRIVER_PRIVATE));
  if (ptDriverSDI->privateData==NULL) {
    printf("ptAllocSDI: alloc ptDriverSDI->privateData failed\n");
    free(ptDriverSDI);
    return NULL;
  }
  memset(ptDriverSDI->privateData, 0, sizeof(PT_DRIVER_PRIVATE));

  if ((ptDriverSDI->privateData->ptSDISem = semBCreate (SEM_Q_PRIORITY, SEM_FULL)) == NULL) {
    perror ("ptAllocSDI: cannot create semaphore");
    free(ptDriverSDI->privateData);
    free(ptDriverSDI);
    return NULL;
  }

  /* Initialisation */
  ptDriverSDI->privateData->ptSerialFd = -1;
  ptDriverSDI->ptVerbose = verbose;
  strcpy(ptDriverSDI->ptDriverName,ptDriverName);

  /* Records driver in the drivers array */
  for (i=0;i<PT_NB_MAX_DRIVERS;i++) {
    if (ptDriversTab[i] == NULL) {
      ptDriversTab[i] = ptDriverSDI;
      ptDriverSDI->privateData->numDriver = i;
      break;
    }
  }
  if (i==PT_NB_MAX_DRIVERS) {
    printf ("ptAllocSDI: too many drivers... reboot recomended !\n");
  }

  /* Records default driver */
  if (ptFirstDriverSDI==NULL)
    ptFirstDriverSDI = ptDriverSDI;
  
  /* Init all data */
  ptDriverSDI->ptResolution = FALSE;
  ptDriverSDI->ptStopDriverFlag = TRUE;
  ptInitSDI(ptDriverSDI);

  return ptDriverSDI;
}

static void
ptInitSDI(PT_DRIVER_STR *ptDriverSDI)
{

  ptDriverSDI->ptDriverInitialized = FALSE;

  ptDriverSDI->privateData->ptMsgFlag = FALSE;
  ptDriverSDI->privateData->ptErr = -1;
  ptDriverSDI->privateData->ptNbWaitedRepliesFifoList = 0;

  /* On reset tout */
  ptResetCmd(PT_ALL, ptDriverSDI);
}

/*
 * ptFreeSDI  -  Cleans up the Internal Data Structure (that makes 'SDI' 
 *               in french) of the driver
 */
static void 
ptFreeSDI(PT_DRIVER_STR *ptDriverSDI) 
{
  int i, numSDI=-1;

  /* Looks for the driver and updates ptFirstDriverSDI */
  if (ptDriverSDI == ptFirstDriverSDI) {
    ptFirstDriverSDI = NULL;
    for (i=0;i<PT_NB_MAX_DRIVERS;i++) {
      if (ptDriversTab[i] != NULL) {
	if (ptDriversTab[i] == ptDriverSDI) 
	  numSDI = i;
	else if (!ptFirstDriverSDI)
	  ptFirstDriverSDI = ptDriversTab[i];
      }
    }
  }

  /* Cleans up */
  semDelete(ptDriverSDI->privateData->ptSDISem);
  free(ptDriverSDI->privateData);
  free(ptDriverSDI);
  ptDriversTab[numSDI] = NULL;
}

static PT_DRIVER_STR *
ptLookForDriver(char *ptDriverName)
{
  int i;
  int tid;
  PT_DRIVER_STR *ptDriverSDI=NULL;

  /* For each recorded driver ... */
  for (i=0;i<PT_NB_MAX_DRIVERS;i++) {
    if (ptDriversTab[i] != NULL) {

      /* if dead, cleans up its data structure */
      if(taskNameToId(ptDriversTab[i]->ptDriverName) == ERROR) {

	printf ("ptCleanOldDrivers: %s seems dead: cleaning SDI\n",
		ptDriversTab[i]->ptDriverName);
	ptFreeSDI(ptDriversTab[i]);
	ptDriversTab[i] = NULL;
      }
      /* else tests if it is the good one */
      else
	if (strcmp(ptDriverName, (ptDriversTab[i])->ptDriverName) == 0) 
	  ptDriverSDI = ptDriversTab[i];
    }
  }

  /* not found in the recorded list */
  if (ptDriverSDI==NULL) {

    /* tests if the process is running */
    if((tid = taskNameToId(ptDriverName)) != ERROR) {

      /* kills it */
	printf ("ptLookForDriver: lonely driver %s deleting task\n",
		ptDriverName);
	taskDelete(tid);
    }
  }

  return ptDriverSDI;
}

/* ====================================================================== *
 *                   FUNCTIONS TO CONTROL THE DRIVER                      *
 * ====================================================================== */


void
ptVerboseSwitch(void)
{
  int i;
  for (i=0;i<PT_NB_MAX_DRIVERS;i++) {
    if (ptDriversTab[i] != NULL) {
      if (ptDriversTab[i]->ptVerbose) {
	ptDriversTab[i]->ptVerbose = FALSE;
      }
      else {
	ptDriversTab[i]->ptVerbose = TRUE;
      }
    }
  }
}


void
ptDisplayDrivers(void)
{
  int i, j;
  PT_DRIVER_PRIVATE *private;

  for(i=0;i<PT_NB_MAX_DRIVERS;i++) {
    if (ptDriversTab[i] != NULL) {
      private = ptDriversTab[i]->privateData;

      printf ("Driver D%d : %s id %p\n",
	      i, ptDriversTab[i]->ptDriverName, ptDriversTab[i]);
      if (!ptDriversTab[i]->ptDriverInitialized) {
	printf ("  not initialized\n");
	continue;
      }
      printf ("   - resolution: %f deg  (%f step/deg)\n"
	      "   - pan [%f %f] tilt [%f %f]\n"
	      "   - stopFlag = %d\n" ,
	      ptDriversTab[i]->ptStep2deg,
	      ptDriversTab[i]->ptDeg2step,
	      PT_STEP_TO_DEG(private->ptDataVal[GET_MIN_POS_PAN].ival, 
			     ptDriversTab[i]),
	      PT_STEP_TO_DEG(private->ptDataVal[GET_MAX_POS_PAN].ival, 
			     ptDriversTab[i]),
	      PT_STEP_TO_DEG(private->ptDataVal[GET_MIN_POS_TILT].ival, 
			     ptDriversTab[i]),
	      PT_STEP_TO_DEG(private->ptDataVal[GET_MAX_POS_TILT].ival, 
			     ptDriversTab[i]),
	      ptDriversTab[i]->ptStopDriverFlag);

      /*==== TAKE SEM ====*/
      if (!ptTakeSDI(ptDriversTab[i], "ptDisplayDrivers")) return;

      ptSpyMsgs(ptDriversTab[i]);

      /* FIFO list of sent commands */
      printf ("   - %d answers waited (in chrono order): ", 
	      private->ptNbWaitedRepliesFifoList);
      for (j=0;j<private->ptNbWaitedRepliesFifoList;j++) {
	printf (" %d ", private->ptWaitedRepliesFifoList[j]);
      }
      printf ("\n");

      /* Status of all on going cmds */
      printf ("   - Status of all on going cmds: ");
      ptSpyAll(ptDriversTab[i]);

      /*==== GIVE SEM ====*/
      ptGiveSDI(ptDriversTab[i], "ptDisplayDrivers");

    }
  }
}


/*
 *  ptResetCmd  --  Resets ALL the status of command 'cmd'
 *                  (all commands if 'cmd == PT_ALL').
 */

void ptResetCmd (int cmd, PT_DRIVER_STR *ptDriverSDI)
{
  int i;

  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /*==== TAKE SEM ====*/
  if (!ptTakeSDI(ptDriverSDI, "ptResetCmd")) {
    printf ("ptResetCmd: cannot take sem\n");
    return;
  }

  /* Resets all command status */
  if (cmd == PT_ALL /* -1 */) {
    ptDriverSDI->privateData->ptNbWaitedRepliesFifoList = 0;
    for (i=0; i<PT_NB_CMD; i++) {
      ptDriverSDI->privateData->ptCmdState[i].replyErr = PT_NO_ERR;
      ptDriverSDI->privateData->ptCmdState[i].replyStatus = PT_NO_CMD;
      ptDriverSDI->privateData->ptCmdState[i].nbWaitedReplies = 0;
      ptDriverSDI->privateData->ptCmdState[i].nbReceivedReplies = 0;
      ptDriverSDI->privateData->ptCmdState[i].nbRepliesToRemove = 0;
    }
  }

  /* Resets status of a given command */
  else if (cmd >= 0 && cmd < PT_NB_CMD) {
    ptDriverSDI->privateData->ptNbWaitedRepliesFifoList -= ptDriverSDI->privateData->ptCmdState[cmd].nbWaitedReplies;
    ptDriverSDI->privateData->ptCmdState[cmd].nbWaitedReplies = 0;
    ptDriverSDI->privateData->ptCmdState[cmd].nbReceivedReplies = 0;
    ptDriverSDI->privateData->ptCmdState[cmd].nbRepliesToRemove = 0;
  }
  /*==== GIVE SEM ====*/
  ptGiveSDI(ptDriverSDI, "ptResetCmd");

}

/*
 *  ptRemoveCmd  --  Removes command 'cmd'
 *                   (ie, reads the answer if already there or flags the 
 *                    command in order to put it out when it will arrive)
 */

BOOL ptRemoveCmd (int cmd, PT_DRIVER_STR *ptDriverSDI)
{
  BOOL status = TRUE;

  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Tests command validity */
  if (!ptTestCmd(cmd, ptDriverSDI))
    return FALSE;

  /*==== TAKE SEM ====*/
  if (!ptTakeSDI(ptDriverSDI, "ptRemoveCmd")) return FALSE;

  /* Cleans up the command if answer is there ... */
  if(ptGetAndUpdateCmdStatusNoProtect(cmd, ptDriverSDI) == PT_WAITING_REPLY) {

    /* ... else indicates that the answer is not pertinent any more */
    if (ptDriverSDI->ptVerbose) {
      printf ("Reply to remove: %d\n", cmd);
    }
    ptDriverSDI->privateData->ptCmdState[cmd].nbRepliesToRemove++;
  }

  /*==== GIVE SEM ====*/
  ptGiveSDI(ptDriverSDI, "ptRemoveCmd");

  return status;
}

/*
 *  ptDriverStop  --  Stops ptDriver
 */

void ptDriverStop(PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Flag to stop */
  ptDriverSDI->ptStopDriverFlag = TRUE;

  /* One last cmd to unlock the task */
  ptSendAsciiCmd("pp", ptDriverSDI);

  if (ptDriverSDI->ptVerbose) {
      printf ("D%d Hope that ptDriver %s does not exist any more\n", 
	      ptDriverSDI->privateData->numDriver, 
	      ptDriverSDI->ptDriverName);
  }
}



/*======================================================================
 *                    EMISSION DES COMMANDES
 *======================================================================*/

/* 
 * ptSendAsciiCmd   -   Sends directly an ASCII string to the PTU
 *
 * Ex: ptSendAsciiCmd "pp10"
 *
 * This command is not interpreted, the answer will be not decoded
 *
 * Retourns : TRUE if OK (else FALSE)
 */
static BOOL ptSendAsciiCmd(char *cmd, PT_DRIVER_STR *ptDriverSDI)
{
  int msgLen;
  char msg[PT_MAX_MSG_LENGTH] = "\0";

  /* Test init */
  if (ptDriverSDI->privateData->ptSerialFd == -1) {
    fprintf (stderr, "ptSend: RS232 line not initialized\n");
    return FALSE;
  }

  /* Builds the cmd */
  strcpy(msg, cmd);
  msgLen = strlen (msg);
  /* Cleans up end of 'msg' */
  while (msg[msgLen] == ' ' || msg[msgLen] == '\n' 
	 || msg[msgLen] == '\r' || msg[msgLen] == '\t') {
    msg[msgLen] = '\0';
    msgLen--;
  }

  /* [command interpretation could be add there] */

  /* Adds ' \n' to indicate the end of the command
     ('\n' not always enough) */
  strcat(msg, "\n");
  msgLen = strlen (msg);  
  if (msgLen > PT_MAX_MSG_LENGTH) {
    fprintf (stderr, "ptSend: Msg %s too long (%d > %d)\n", 
	    msg, msgLen, PT_MAX_MSG_LENGTH);
    return FALSE;
  }

  /* Sends string */
/* #define DONTWRITE */
#ifndef DONTWRITE
  write (ptDriverSDI->privateData->ptSerialFd, msg, msgLen);
  if (ptDriverSDI->ptVerbose)
    printf ("write: -%s-\n", msg); 
#else
  printf ("false write: -%s -(len=%d fd=%d)\n", msg, msgLen, ptDriverSDI->privateData->ptSerialFd);
#endif

  return TRUE;
}

/* 
 * ptSendMsg   -   Sends an ascii string
 *
 *
 * Retourns TRUE if OK (else FALSE)
 */
BOOL ptSendMsg(char *msg, PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Enregistre la commande ascii */
  if (!ptAddWaitedReplies(PT_ASCII, ptDriverSDI))
    return FALSE;

  /* Envoie */
  ptSendAsciiCmd(msg, ptDriverSDI);
  
  return TRUE;
}

/* 
 * ptSendBasicCmd   -   Sends one of the enum PT_CMD commands (see ptDriver.h)
 *
 * The command is NOT recorded and the answer will be not decoded
 *
 * Parameters: 
 *   <cmd>   : send command (a PT_CMD)
 *   <ival>  : command parameter if any (int)
 *
 * Ex: ptSendBasicCmd(GET_POS_ABS_PAN, 0, TRUE, NULL);
 *
 * Retourns TRUE if OK (else FALSE)
 */
BOOL ptSendBasicCmd(PT_CMD cmd, int ival, PT_DRIVER_STR *ptDriverSDI)
{
  char bufcmd[PT_MAX_MSG_LENGTH];

  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* XXXXX Protection from reset for debug purpose */
  if (cmd == PT_RESET)
    taskDelay(2);
  if (ptDriverSDI->privateData->ptCmdState[PT_RESET].nbWaitedReplies != 0)
    return TRUE;
  /* XXXXX */

  /* Test command validity */
  if (!ptTestCmd(cmd, ptDriverSDI))
    return FALSE;

  /* If it is a constant that we already have, just returns it */
  if (ptDriverSDI->privateData->ptCmdState[cmd].replyStatus == PT_CONST_OK)
    return TRUE;

  /* message construction */
  switch (ptCmdDesc[cmd].inputDataType) {
  case PT_NO_DATA:
    snprintf (bufcmd, sizeof(bufcmd), "%s", ptCmdDesc[cmd].inputFormat);
    break;
  case PT_INT_DATA:
    snprintf (bufcmd, sizeof(bufcmd), ptCmdDesc[cmd].inputFormat, ival);
    ptDriverSDI->privateData->ptDataVal[cmd].ival = ival;
    break;
  default:
    fprintf (stderr, "ptSendCmd: not handled, sorry !\n");
  }

  /* Sends message */
  ptSendAsciiCmd(bufcmd, ptDriverSDI);

  /* Acts like if we have an acknowlegment */
  return TRUE;
}

/* 
 * ptSendCmd   -   Sends one of enum PT_CMD command (see ptDriver.h)
 *
 * The command is recoreded and the answer will be decoded
 *
 * Parametres: 
 *   <cmd>   : send command (PT_CMD)
 *   <ival>  : command parameter if any (int)
 *
 * Ex: ptSendCmd(GET_POS_ABS_PAN, 0, TRUE, NULL);
 *
 * Retourns TRUE if OK (or FALSE)
 */
BOOL ptSendCmd(PT_CMD cmd, int ival, PT_DRIVER_STR *ptDriverSDI)
{
  char bufcmd[PT_MAX_MSG_LENGTH];

  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* XXXXX Protection from reset for debug purpose */
  if (cmd == PT_RESET)
    taskDelay(2);
  if (ptDriverSDI->privateData->ptCmdState[PT_RESET].nbWaitedReplies != 0)
    return TRUE;
  /* XXXXX */

  /* Test command validity */
  if (!ptTestCmd(cmd, ptDriverSDI))
    return FALSE;

  /* If it is a constant that we already have, just returns it */
  if (ptDriverSDI->privateData->ptCmdState[cmd].replyStatus == PT_CONST_OK)
    return TRUE;

  /* message construction */
  switch (ptCmdDesc[cmd].inputDataType) {
  case PT_NO_DATA:
    sprintf (bufcmd, ptCmdDesc[cmd].inputFormat);
    break;
  case PT_INT_DATA:
    sprintf (bufcmd, ptCmdDesc[cmd].inputFormat, ival);
    ptDriverSDI->privateData->ptDataVal[cmd].ival = ival;
    break;
  default:
    fprintf (stderr, "ptSendCmd: not handled, sorry !\n");
  }

  /* Records command */
  if (ptCmdDesc[cmd].outputDataType != PT_NO_ANSWER)
    if (!ptAddWaitedReplies(cmd, ptDriverSDI))
      return FALSE;

  /* Sends message */
  ptSendAsciiCmd(bufcmd, ptDriverSDI);

  /* Acts like if we have an acknowlegment */
  return TRUE;
}

BOOL ptSendCmdRad(PT_CMD cmd, double val, PT_DRIVER_STR *ptDriverSDI)
{
  int ival;
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;
  if (!ptDriverSDI->ptResolution) {
    fprintf (stderr, "ptResolution not initialized\n");
    return FALSE;
  }
  ival = PT_RAD_TO_STEP(val,ptDriverSDI);
  return ptSendCmd(cmd, ival, ptDriverSDI);
}

BOOL ptSendCmdDeg(PT_CMD cmd, double val, PT_DRIVER_STR *ptDriverSDI)
{
  int ival;
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;
  if (!ptDriverSDI->ptResolution) {
    fprintf (stderr, "ptResolution not initialized\n");
    return FALSE;
  }
  ival = PT_DEG_TO_STEP(val,ptDriverSDI);
  return ptSendCmd(cmd, ival, ptDriverSDI);
}

void ptPrintConst(PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  printf ("ptResolution %f\n", ptDriverSDI->ptResolution);
  printf ("ptStep2rad %f\n", ptDriverSDI->ptStep2rad);
  printf ("ptRad2step %f\n", ptDriverSDI->ptRad2step);
  printf ("ptStep2deg %f\n", ptDriverSDI->ptStep2deg);
  printf ("ptDeg2step %f\n", ptDriverSDI->ptDeg2step);
}
/* ====================================================================== *
 *    FUNCTIONS TO READ DATA RESULTING FROM INTERPRETED COMMAND           *
 * ====================================================================== *
 *  The following functions reset the command status to PT_NO_CMD         *
 * ====================================================================== */

/*
 * ptGetStatus  -  gets command status. if the answer is there, resets it
 *                 to PT_NO_CMD
 */

PT_CMD_STATUS ptGetStatus(PT_CMD cmd, PT_ERR *err, PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Tests cmd */
  if (!ptTestCmd(cmd, ptDriverSDI))
    return PT_CMD_UNKNOWN;

  *err = ptDriverSDI->privateData->ptCmdState[cmd].replyErr;

  /* Returns status (resets it eventually) */
  return ptGetAndUpdateCmdStatus (cmd, ptDriverSDI);
}

/*
 * ptGetVal  -  gets status and value, 
 *              if the answer has arrived :
 *                    . resets status to PT_NO_CMD,
 *                    . returns eventually the value (if *val != NULL)
 */

PT_CMD_STATUS ptGetInt(PT_CMD cmd, int *val, PT_ERR *err, 
		       PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Tests cmd */
  if (!ptTestCmd(cmd, ptDriverSDI))
    return PT_CMD_UNKNOWN;

  /* Reads value (independently of the status ! ) */
  if (val != NULL) {
    if (ptCmdDesc[cmd].outputDataType == PT_INT_DATA)
      *val = ptDriverSDI->privateData->ptDataVal[cmd].ival;
    else if (ptCmdDesc[cmd].outputDataType != PT_NO_DATA) {
      fprintf (stderr, "ptGetStepVal: unvalid output val type\n");
      return PT_CMD_UNKNOWN;
    }
  }

  *err = ptDriverSDI->privateData->ptCmdState[cmd].replyErr;

  /* Returns the status (and eventually resets it) */
  return ptGetAndUpdateCmdStatus (cmd, ptDriverSDI);
}

PT_CMD_STATUS ptGetDouble(PT_CMD cmd, double *val, PT_ERR *err, 
			  PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Tests cmd */
  if (!ptTestCmd(cmd,ptDriverSDI))
    return PT_CMD_UNKNOWN;

  /* Reads the value (what ever the status is ! ) */
  if (val != NULL) {

    if (ptCmdDesc[cmd].outputDataType == PT_DOUBLE_DATA_S) {
      *val = ptDriverSDI->privateData->ptDataVal[cmd].dval;
    }
    else {
      fprintf (stderr, "ptGetDouble: unvalid output val type\n");
      return PT_CMD_UNKNOWN;
    }

  } /* val != NULL */

  *err = ptDriverSDI->privateData->ptCmdState[cmd].replyErr;
  /* Returns the status (and eventually resets it) */
  return ptGetAndUpdateCmdStatus(cmd,ptDriverSDI);
}

PT_CMD_STATUS ptGetDeg(PT_CMD cmd, double *val, PT_ERR *err, 
			  PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Tests cmd */
  if (!ptTestCmd(cmd,ptDriverSDI))
    return PT_CMD_UNKNOWN;

  /* Reads the value (what ever the status is ! ) */
  if (val != NULL) {

    /* Converts the integer to a double in degrees */
    if (ptCmdDesc[cmd].outputDataType == PT_INT_DATA) {
      if (ptDriverSDI->ptResolution == 1) {
	fprintf (stderr, "ptGetDouble: ptResolution not init: cannot convert to rad\n");
	return PT_CMD_UNKNOWN;
      }
      else {
	*val = PT_STEP_TO_DEG(ptDriverSDI->privateData->ptDataVal[cmd].ival,
			      ptDriverSDI);
      }
    }
    
    /* Quesaco ? */
    else {
      fprintf (stderr, "ptGetDeg: unvalid output val type\n");
      return PT_CMD_UNKNOWN;
    }

  } /* val != NULL */

  *err = ptDriverSDI->privateData->ptCmdState[cmd].replyErr;

  /* Returns the status (and eventually resets it) */
  return ptGetAndUpdateCmdStatus (cmd,ptDriverSDI);
}

PT_CMD_STATUS ptGetRad(PT_CMD cmd, double *val, PT_ERR *err, 
			  PT_DRIVER_STR *ptDriverSDI)
{
  PT_CMD_STATUS status;
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;
  status = ptGetDeg(cmd, val, err, ptDriverSDI);
  *val = DEG_TO_RAD(*val);
  return status;
}

PT_CMD_STATUS ptGetString(PT_CMD cmd, char *val, PT_ERR *err, 
			  PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Tests cmd */
  if (!ptTestCmd(cmd,ptDriverSDI))
    return PT_CMD_UNKNOWN;

  /* Reads the value (what ever the status is ! ) */
  if (val != NULL) {
    if (ptCmdDesc[cmd].outputDataType == PT_STRING_DATA)
      strcpy(val, ptDriverSDI->privateData->ptDataVal[cmd].sval);
    else {
      fprintf (stderr, "ptGetString: unvalid output val type\n");
      return PT_CMD_UNKNOWN;
    }
  }

  *err = ptDriverSDI->privateData->ptCmdState[cmd].replyErr;
  /* Returns the status (and eventually resets it) */
  return ptGetAndUpdateCmdStatus (cmd, ptDriverSDI);
}


PT_CMD_STATUS ptGetStringCode(PT_CMD cmd, int *val, PT_ERR *err, 
			      PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Tests cmd */
  if (!ptTestCmd(cmd, ptDriverSDI))
    return PT_CMD_UNKNOWN;

  /* Reads the value (what ever the status is ! ) */
  if (val != NULL) {
    if (ptCmdDesc[cmd].outputDataType == PT_STRING_DATA) {
      *val = ptDecodeString(ptDriverSDI->privateData->ptDataVal[cmd].sval, ptDriverSDI);
    }
    else if (ptCmdDesc[cmd].outputDataType != PT_NO_DATA) {
      fprintf (stderr, "ptGetStringCode: unvalid output val type\n");
      return PT_CMD_UNKNOWN;
    }
  }
  *err = ptDriverSDI->privateData->ptCmdState[cmd].replyErr;

  /* Returns the status (and eventually resets it) */
  return ptGetAndUpdateCmdStatus (cmd, ptDriverSDI);
}

static PT_STRINGS_DECODE ptDecodeString(char *string,
					PT_DRIVER_STR *ptDriverSDI)
{
  int i;

  for (i=0; i<PT_NB_STRINGS_DECODE; i++) 
    if (strcmp(string, ptStringsDesc[i].format) == 0)
      return i;

  return -1;
}


PT_CMD_STATUS ptGetPrint(PT_CMD cmd, PT_DRIVER_STR *ptDriverSDI) 
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Prints command status */
  if (ptSpyCmd(cmd, ptDriverSDI) == PT_CMD_UNKNOWN)
    return PT_CMD_UNKNOWN;

  /* Returns the status (and eventually resets it) */
  return ptGetAndUpdateCmdStatus (cmd, ptDriverSDI);
}

/* ====================================================================== *
 *           FUNCTIONS TO SEND COMMAND AND TO READ RESULTS                *
 * ====================================================================== *
 *  The following functions reset the command status to PT_NO_CMD         *
 * ====================================================================== */

PT_CMD_STATUS ptSendAndGet(PT_CMD cmd, int *ival, BOOL block, 
			   PT_ERR *err, PT_DRIVER_STR *ptDriverSDI)
{
  PT_CMD_STATUS status;

  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Is there already an answer ? */
  switch(ptCmdDesc[cmd].outputDataType) {
  case PT_INT_DATA:
    status = ptGetInt(cmd, ival, err, ptDriverSDI);
    break;
  case PT_STRING_DATA:
    status = ptGetStringCode(cmd, ival, err, ptDriverSDI);
    break;
  case PT_NO_DATA:
  case PT_UNKNOWN_DATA:
  case PT_STRING_NO_DATA:
    status = ptGetStatus(cmd, err, ptDriverSDI);
    break;
  case PT_NO_ANSWER:
    status = PT_NO_CMD;
    break;
  case PT_DOUBLE_DATA_S:
  default:
    return PT_CMD_UNKNOWN;
  }

  /* Sends command */
  switch (status) {
  case PT_NO_CMD:
    if (!ptSendCmd(cmd, *ival, ptDriverSDI))
      return PT_CMD_UNKNOWN;
    break;
  case PT_REPLY_OK:
  case PT_REPLY_ERROR:
  case PT_CONST_OK:
  case PT_CMD_UNKNOWN:
  case PT_REPLY_TIMEOUT:
    return status;
  case PT_WAITING_REPLY:
    if (!block)
      return status;
  }

  do {
    switch(ptCmdDesc[cmd].outputDataType) {
    case PT_INT_DATA:
      status = ptGetInt(cmd, ival, err, ptDriverSDI);
      break;
    case PT_STRING_DATA:
      status = ptGetStringCode(cmd, ival, err, ptDriverSDI);
      break;
    case PT_UNKNOWN_DATA:
    case PT_NO_DATA:
    case PT_STRING_NO_DATA:
      status = ptGetStatus(cmd, err, ptDriverSDI);
      break;
    case PT_NO_ANSWER:
      status = PT_REPLY_OK;
      break;
    case PT_DOUBLE_DATA_S:
      return PT_CMD_UNKNOWN;
    }

  } while (block && status == PT_WAITING_REPLY && taskDelay(PT_NB_TICS_PER_CYCLE_IN_BLOCK_MODE) == OK);
   
  return status;
}

PT_CMD_STATUS ptGetSendAndGet(PT_CMD cmd, int *ival, 
			      PT_ERR *err, PT_DRIVER_STR *ptDriverSDI)
{
  PT_CMD_STATUS status;

  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Is there already an answer ? */
  switch(ptCmdDesc[cmd].outputDataType) {
  case PT_INT_DATA:
    status = ptGetInt(cmd, ival, err, ptDriverSDI);
    break;
  case PT_STRING_DATA:
    status = ptGetStringCode(cmd, ival, err, ptDriverSDI);
    break;
  case PT_NO_DATA:
  case PT_UNKNOWN_DATA:
  case PT_STRING_NO_DATA:
    status = ptGetStatus(cmd, err, ptDriverSDI);
    break;
  case PT_NO_ANSWER:
    status = PT_NO_CMD;
    break;
  case PT_DOUBLE_DATA_S:
  default:
    return PT_CMD_UNKNOWN;
  }

  /* Sends command */
  switch (status) {
  case PT_WAITING_REPLY:
    if (ptDriverSDI->ptVerbose)
      printf ("D%d rm cmd %d\n", ptDriverSDI->privateData->numDriver, cmd);
    ptRemoveCmd(cmd, ptDriverSDI);

    if (!ptSendCmd(cmd, *ival, ptDriverSDI))
      return PT_CMD_UNKNOWN;
    break;

  case PT_NO_CMD:
    if (!ptSendCmd(cmd, *ival, ptDriverSDI))
      return PT_CMD_UNKNOWN;
    break;

  case PT_REPLY_OK:
  case PT_REPLY_ERROR:
  case PT_CONST_OK:
  case PT_CMD_UNKNOWN:
  case PT_REPLY_TIMEOUT:
    if (ptDriverSDI->ptVerbose)
      printf ("D%d status %d\n", ptDriverSDI->privateData->numDriver, status);
    return status;
  }

  /* Reads answer if any */

  taskDelay(2);

  switch(ptCmdDesc[cmd].outputDataType) {
  case PT_INT_DATA:
    status = ptGetInt(cmd, ival, err, ptDriverSDI);
    break;
  case PT_STRING_DATA:
    status = ptGetStringCode(cmd, ival, err, ptDriverSDI);
    break;
  case PT_NO_DATA:
  case PT_UNKNOWN_DATA:
  case PT_STRING_NO_DATA:
    status = ptGetStatus(cmd, err, ptDriverSDI);
      break;
  case PT_NO_ANSWER:
    status = PT_REPLY_OK;
    break;
  case PT_DOUBLE_DATA_S:
    return PT_CMD_UNKNOWN;
  }
   
  return status;
}

PT_CMD_STATUS ptGetSendAndGetRad(PT_CMD cmd, double *val, 
				 PT_ERR *err, PT_DRIVER_STR *ptDriverSDI)
{
  int ival;
  PT_CMD_STATUS status;
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;
  if (!ptDriverSDI->ptResolution) {
    fprintf (stderr, "D%d ptResolution not initialized\n", 
	     ptDriverSDI->privateData->numDriver);
    return FALSE;
  }
  ival = PT_RAD_TO_STEP(*val,ptDriverSDI);
  status = ptGetSendAndGet(cmd, &ival, err, ptDriverSDI);
  *val = PT_STEP_TO_RAD(ival,ptDriverSDI);
  return status;
}

PT_CMD_STATUS ptSendAndGetRad(PT_CMD cmd, double *val, BOOL block, 
			      PT_ERR *err, PT_DRIVER_STR *ptDriverSDI)
{
  int ival;
  PT_CMD_STATUS status;
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;
  if (!ptDriverSDI->ptResolution) {
    fprintf (stderr, "D%d ptResolution not initialized\n", 
	     ptDriverSDI->privateData->numDriver);
    return FALSE;
  }
  ival = PT_RAD_TO_STEP(*val,ptDriverSDI);
  status = ptSendAndGet(cmd, &ival, block, err, ptDriverSDI);
  *val = PT_STEP_TO_RAD(ival,ptDriverSDI);
  return status;
}

PT_CMD_STATUS ptSendAndGetDeg(PT_CMD cmd, double *val, BOOL block, 
			      PT_ERR *err, PT_DRIVER_STR *ptDriverSDI)
{
  int ival;
  PT_CMD_STATUS status;
  if (!ptDriverSDI->ptResolution) {
    fprintf (stderr, "D%d ptResolution not initialized\n", 
	     ptDriverSDI->privateData->numDriver);
    return FALSE;
  }
  ival = PT_DEG_TO_STEP(*val,ptDriverSDI);
  status = ptSendAndGet(cmd, &ival, block, err, ptDriverSDI);
  *val = PT_STEP_TO_DEG(ival,ptDriverSDI);
  return status;
}

PT_CMD_STATUS ptSendAndGetString(PT_CMD cmd, char *val, 
				 BOOL block, PT_ERR *err, 
				 PT_DRIVER_STR *ptDriverSDI)
{
  PT_CMD_STATUS status;
  int pipo=0;

  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Is there already an answer ? */
  status = ptGetString(cmd, val, err, ptDriverSDI);

  /* Sends command */
  switch (status) {
  case PT_NO_CMD:
    if (!ptSendCmd(cmd, pipo, ptDriverSDI))
      return PT_CMD_UNKNOWN;
    break;
  case PT_REPLY_OK:
  case PT_REPLY_ERROR:
  case PT_CONST_OK:
  case PT_CMD_UNKNOWN:
  case PT_REPLY_TIMEOUT:
    return status;
  case PT_WAITING_REPLY:
    if (!block)
      return status;
  }

  do {
    status = ptGetString(cmd, val, err, ptDriverSDI);

  } while (block && status == PT_WAITING_REPLY && taskDelay(PT_NB_TICS_PER_CYCLE_IN_BLOCK_MODE)==OK);
   
  return status;
}


/* ====================================================================== *
 *               FUNCTIONS TO SEND COMMAND ET PRINT RESULTS               *
 * ====================================================================== */

void ptPrint(FILE *out, char *msg, PT_CMD cmd, 
	     PT_CMD_STATUS status, PT_ERR err, PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;
  if (msg != NULL)
    fprintf(out, "D%d %s: ", ptDriverSDI->privateData->numDriver, msg);
  if (cmd>=0 && cmd<PT_NB_CMD)
    fprintf(out, "cmd %s ", ptCmdDesc[cmd].name);
  if (status>=0 && status <sizeof(ptCmdStatusStrings)/sizeof(char *))
    fprintf(out, "status %s ", ptCmdStatusStrings[status]);
  if (err>=0 && err <PT_NB_ERR)
    fprintf(out, "err %s ", ptErrDesc[err].name);
  fprintf(out, "\n");
}

BOOL ptSendAndGetPrint(PT_CMD cmd, int ival, PT_DRIVER_STR *ptDriverSDI)
{
  PT_ERR err;
  PT_CMD_STATUS status;

  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  /* Sends command */
  if (!ptSendCmd(cmd, ival, ptDriverSDI))
    return FALSE;

  /* Wait for answer */
  while ((status = ptGetStatus(cmd, &err, ptDriverSDI)) == PT_WAITING_REPLY) {
    taskDelay(PT_NB_TICS_PER_CYCLE_IN_BLOCK_MODE);
  }

  ptGetPrint(cmd, ptDriverSDI);
  printf ("   New status : %s\n", ptCmdStatusStrings[status]);
  
  if (status != PT_CONST_OK && status != PT_REPLY_OK) 
    return FALSE;

  return TRUE;
}

BOOL ptSendCmdAndGetPrintRad(PT_CMD cmd, double val, 
			     PT_DRIVER_STR *ptDriverSDI)
{
  int ival;
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  if (!ptDriverSDI->ptResolution) {
    fprintf (stderr, "D%d ptResolution not initialized\n", 
	     ptDriverSDI->privateData->numDriver);
    return FALSE;
  }
  ival = PT_RAD_TO_STEP(val,ptDriverSDI);
  return ptSendAndGetPrint (cmd, ival, ptDriverSDI);
}

BOOL ptSendCmdAndGetPrintDeg(PT_CMD cmd, double val, 
			     PT_DRIVER_STR *ptDriverSDI)
{
  int ival;
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  if (!ptDriverSDI->ptResolution) {
    fprintf (stderr, "ptResolution not initialized\n");
    return FALSE;
  }
  ival = PT_DEG_TO_STEP(val,ptDriverSDI);
  return ptSendAndGetPrint (cmd, ival, ptDriverSDI);
}

/* ====================================================================== *
 *                       FUNCTIONS TO READ RAW MESSAGES                   *
 * ====================================================================== */

/*
 * ptReadMsg  -  Reads last message (ptDriverSDI->privateData->ptMsg) in *buf
 *               and removes ptDriverSDI->privateData->ptMsg
 * 
 * Returns FALSE if no msg
 */

BOOL ptReadMsg(char *buf, PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  strcpy(buf, ptDriverSDI->privateData->ptMsg);
  if (!ptDriverSDI->privateData->ptMsgFlag) {
    fprintf (stderr, "D%d NO REMAINING MSG\n", ptDriverSDI->privateData->numDriver);
    return FALSE;
  }
  ptDriverSDI->privateData->ptMsgFlag = FALSE;
  return TRUE;
}


/*
 * ptPrintMsg  -  like ptReadMsg but prints msg and does not returns buf
 */
BOOL ptPrintMsg(PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  if (!ptDriverSDI->privateData->ptMsgFlag) {
    printf ("D%d NO REMAINING MSG (previous: %s)\n", 
	    ptDriverSDI->privateData->numDriver, ptDriverSDI->privateData->ptMsg);
    return FALSE;
  }
  printf("D%d REMAINING MSG: %s\n", ptDriverSDI->privateData->numDriver, 
	 ptDriverSDI->privateData->ptMsg);
  ptDriverSDI->privateData->ptMsgFlag = FALSE;
  return TRUE;
}

/* ====================================================================== *
 *                    FUNCTIONS TO SPY (no sides effects !)               *
 * ====================================================================== */

void ptSpyMsgs(PT_DRIVER_STR *ptDriverSDI)
{
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

  printf ("D%d (ptMsgFlag %d) ptMsg %s\n", ptDriverSDI->privateData->numDriver, 
	  ptDriverSDI->privateData->ptMsgFlag, ptDriverSDI->privateData->ptMsg);
  printf ("D%d (ptErrMsg %s)  ptErr %d %s\n", ptDriverSDI->privateData->numDriver, 
	  ptDriverSDI->privateData->ptErrMsg, ptDriverSDI->privateData->ptErr, 
	  ptDriverSDI->privateData->ptErr>=0 && ptDriverSDI->privateData->ptErr<PT_NB_ERR ? ptErrDesc[ptDriverSDI->privateData->ptErr].name : "?");
}

PT_CMD_STATUS ptSpyCmd(PT_CMD cmd, PT_DRIVER_STR *ptDriverSDI)
{
  PT_CMD_STATUS status; 
  int code;
  PT_ERR ptErr;

   if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;

 /* Tests cmd */
  if (!ptTestCmd(cmd,ptDriverSDI))
    return PT_CMD_UNKNOWN;
  
  /* Reads status */
  status = ptGetCmdStatus(cmd,ptDriverSDI);
  ptErr = ptDriverSDI->privateData->ptCmdState[cmd].replyErr;

  printf ("   D%d %s %d:  status %s (err %s)\n", ptDriverSDI->privateData->numDriver, 
	  ptCmdDesc[cmd].name,  cmd,
	  ptCmdStatusStrings[status], 
	  status != PT_REPLY_ERROR ? "none" :
	  ptErr>=0 && ptErr<PT_NB_ERR ? ptErrDesc[ptErr].name : "?");


  switch (ptCmdDesc[cmd].outputDataType) {
  case PT_UNKNOWN_DATA:
  case PT_NO_DATA:
  case PT_STRING_NO_DATA:
    ptSpyCmdInput(cmd, ptDriverSDI);
    break;
  case PT_INT_DATA:
    if (ptDriverSDI->privateData->ptCmdState[cmd].replyErr == PT_UNKNOWN_ERR)
      break;
    if (ptCmdDesc[cmd].inputDataType != PT_NO_DATA) 
      printf ("       in: lost ! \n");
    printf ("        out: %d steps / %f rad / %f deg\n",
	    ptDriverSDI->privateData->ptDataVal[cmd].ival,
	    PT_STEP_TO_RAD(ptDriverSDI->privateData->ptDataVal[cmd].ival,
			   ptDriverSDI),
	    PT_STEP_TO_DEG(ptDriverSDI->privateData->ptDataVal[cmd].ival,
			   ptDriverSDI));
    break;
  case PT_DOUBLE_DATA_S:
    if (ptDriverSDI->privateData->ptCmdState[cmd].replyErr == PT_UNKNOWN_ERR)
      break;
    if (ptCmdDesc[cmd].inputDataType != PT_NO_DATA) 
      printf ("       in: lost ! \n");
    printf ("        out: %f\n", ptDriverSDI->privateData->ptDataVal[cmd].dval);
    break;
  case PT_STRING_DATA:
    if (ptDriverSDI->privateData->ptCmdState[cmd].replyErr == PT_UNKNOWN_ERR)
      break;
    code = ptDecodeString(ptDriverSDI->privateData->ptDataVal[cmd].sval, ptDriverSDI);
    if (ptCmdDesc[cmd].inputDataType != PT_NO_DATA) 
      printf ("       in: lost ! \n");
    if (code != -1)
      printf ("        out: %s\n", ptStringsDesc[code].format);
    else
      printf ("        out: %s (unknown)\n", ptDriverSDI->privateData->ptDataVal[cmd].sval);
    break;
  case PT_NO_ANSWER:
    break;
  }

  return status;
}

static void ptSpyCmdInput(int cmd, PT_DRIVER_STR *ptDriverSDI) 
{
  int code;

  switch (ptCmdDesc[cmd].inputDataType) {
  case PT_UNKNOWN_DATA:
  case PT_NO_DATA:  
  case PT_STRING_NO_DATA:
  case PT_NO_ANSWER:
    printf ("        in: no data\n");
    break;
  case PT_INT_DATA:
    printf ("        in: %d / %f rad / %f deg\n",
	    ptDriverSDI->privateData->ptDataVal[cmd].ival,
	    PT_STEP_TO_RAD(ptDriverSDI->privateData->ptDataVal[cmd].ival,
			   ptDriverSDI),
	    PT_STEP_TO_DEG(ptDriverSDI->privateData->ptDataVal[cmd].ival,
			   ptDriverSDI));
    break;
  case PT_DOUBLE_DATA_S:
    printf ("        in: %f\n", ptDriverSDI->privateData->ptDataVal[cmd].dval);
    break;
  case PT_STRING_DATA:
    code = ptDecodeString(ptDriverSDI->privateData->ptDataVal[cmd].sval, ptDriverSDI);
    if (code != -1)
      printf ("        in: %s\n", ptStringsDesc[code].format);
    else
      printf ("        in: %s (unknown)\n", ptDriverSDI->privateData->ptDataVal[cmd].sval);

    break;
  }
}
void ptSpyAll(PT_DRIVER_STR *ptDriverSDI)
{
  int i;
  PT_CMD_STATUS status;

  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;
  for (i=0; i<PT_NB_CMD; i++) {
    status = ptGetCmdStatus(i, ptDriverSDI);
    if (status != PT_NO_CMD && status != PT_CONST_OK)
      ptSpyCmd(i, ptDriverSDI);
  }
}
void ptSpyVeryAll(PT_DRIVER_STR *ptDriverSDI)
{
  int i;
  if (ptDriverSDI==NULL)
    ptDriverSDI = ptFirstDriverSDI;
  for (i=0; i<PT_NB_CMD; i++)
    ptSpyCmd(i, ptDriverSDI);
}

/* ====================================================================== *
 *                      LOCAL FUNCTIONS                                   *
 * ====================================================================== */

/*
 * ptGetCmdStatus  -  Computes command status
 *                    No side effect
 */
static PT_CMD_STATUS ptGetCmdStatus (int cmd, PT_DRIVER_STR *ptDriverSDI)
{
  /* It is a  constante that we already have */
  if (ptDriverSDI->privateData->ptCmdState[cmd].replyStatus == PT_CONST_OK) 
    return PT_CONST_OK;

  /* No on going command */
  if (!ptDriverSDI->privateData->ptCmdState[cmd].nbWaitedReplies && 
      !ptDriverSDI->privateData->ptCmdState[cmd].nbReceivedReplies)
    return PT_NO_CMD;

  /* Answer(s) waiting to be read by the client */
  if (ptDriverSDI->privateData->ptCmdState[cmd].nbReceivedReplies)
    return ptDriverSDI->privateData->ptCmdState[cmd].replyStatus;

  /* Waiting for answer */
  if (ptDriverSDI->privateData->ptCmdState[cmd].nbWaitedReplies) {

    /* Timeout */
    if (ptDriverSDI->privateData->ptCmdState[cmd].timeout == 0)
      return PT_REPLY_TIMEOUT;
    /* Wait */
    return PT_WAITING_REPLY;
  }

  return PT_CMD_UNKNOWN;
} 

/*
 * ptGetAndUpdateCmdStatus - computes command status and resets it if
 *                           the answer is there
 */
static PT_CMD_STATUS ptGetAndUpdateCmdStatus (int cmd, 
					      PT_DRIVER_STR *ptDriverSDI)
{
  PT_CMD_STATUS status;

  /*==== TAKE SEM ====*/
  if (!ptTakeSDI(ptDriverSDI, "ptGetAndUpdateCmdStatus")) 
    return PT_REPLY_ERROR;

  status = ptGetAndUpdateCmdStatusNoProtect(cmd, ptDriverSDI);

  /*==== GIVE SEM ====*/
  ptGiveSDI(ptDriverSDI, "ptGetAndUpdateCmdStatus");

  return status;
} 
static PT_CMD_STATUS ptGetAndUpdateCmdStatusNoProtect (int cmd, 
						       PT_DRIVER_STR *ptDriverSDI)
{
  PT_CMD_STATUS status;

  /* Reads  command status */
  status = ptGetCmdStatus(cmd,ptDriverSDI);

  /* Answer is there */
  if (status == PT_REPLY_OK || status == PT_REPLY_ERROR) {
    ptDriverSDI->privateData->ptCmdState[cmd].nbReceivedReplies--;
    return status;
  }

  /* No need to wait any more */
  if (status == PT_REPLY_TIMEOUT) {
    printf ("Reply TIMEOUT: %d\n", cmd);
    /* We suppose that the answer will never arrived ...
       ptDriverSDI->privateData->ptCmdState[cmd].nbRepliesToRemove++; */
    ptDriverSDI->privateData->ptCmdState[cmd].replyErr = PT_TIMEOUT_ERR;
    return status;
  }

  /* Still waiting */
  if (ptDriverSDI->privateData->ptCmdState[cmd].timeout)
    ptDriverSDI->privateData->ptCmdState[cmd].timeout--;

  return status;
} 

/*
 * ptTestCmd  -  tests command validity
 */
static BOOL ptTestCmd(int cmd, PT_DRIVER_STR *ptDriverSDI)
{
  if (cmd < 0 || cmd >= PT_NB_CMD) {
    fprintf (stderr, "D%d Unknown command %d\n", 
	     ptDriverSDI->privateData->numDriver, cmd);
    return FALSE;
  }
  return TRUE;
}

/* ====================================================================== *
 *                 MANAGEMENT OF THE LIST OF WAITING COMMANDS             *
 * ====================================================================== */

/*
 * ptAddWaitedReplies - 
 */

static BOOL ptAddWaitedReplies(PT_CMD cmd, PT_DRIVER_STR *ptDriverSDI)
{

  /*==== TAKE SEM ====*/
  if (!ptTakeSDI(ptDriverSDI, "ptAddWaitedReplies")) return FALSE;

  /* Add in fifo list */
  if (ptDriverSDI->privateData->ptNbWaitedRepliesFifoList == PT_MAX_WAITED_REPLIES-1) {
    /* XXXX How this situation can happen ? ... */
    printf ("D%d ptAddWaitedReplies: too many requests -> RESET ALL !!!\n",
	    ptDriverSDI->privateData->numDriver);
    /*==== GIVE SEM ====*/
    ptGiveSDI(ptDriverSDI, "ptAddWaitedReplies");
    ptResetCmd(PT_ALL, ptDriverSDI);
    return FALSE;
  }

  /* Add in fifo list  */
  ptDriverSDI->privateData->ptWaitedRepliesFifoList[ptDriverSDI->privateData->ptNbWaitedRepliesFifoList] = cmd;
  (ptDriverSDI->privateData->ptNbWaitedRepliesFifoList)++;

  
  (ptDriverSDI->privateData->ptCmdState[cmd].nbWaitedReplies)++;

  /* To be done : timeout value should perhaps depend on the command 
     type XXXX */ 
  ptDriverSDI->privateData->ptCmdState[cmd].timeout = PT_TIMEOUT;

  /*==== GIVE SEM ====*/
  ptGiveSDI(ptDriverSDI, "ptAddWaitedReplies");

  return TRUE;
}



